import transcode, os, socket, re,zlib
from Cheetah.Template import Template
from plugin import Plugin
from urllib import unquote_plus, quote, unquote
from urlparse import urlparse
from xml.sax.saxutils import escape
from lrucache import LRUCache
from UserDict import DictMixin
from datetime import datetime, timedelta
import urllib
from xml.dom import minidom
import config
import time

SCRIPTDIR = os.path.dirname(__file__)
GETSERIESID_URL = '/GetSeries.php?'
GETEPISODEID_URL = '/GetEpisodes.php?'
GETEPISODEINFO_URL = '/EpisodeUpdates.php?'

CLASS_NAME = 'Video'

debug = config.getDebug()
hack83 = config.getHack83()
#mirroring site is broken so set to empty to disable it
#mirrorURL = self.getMirrorURL()
mirrorURL = ""
#Set the default cache time to force reload
dvrLastCacheTime = []
dvrLastCacheTime.append(datetime.now() - timedelta(hours=2))
dvarchiveURL = config.get("ReplayTV","dvarchiveURL")
transcodeEnabled = config.get("ReplayTV","transcode")
dvrShowCache = {}

def debug_write(data):
    if debug:
        debug_out = []
        debug_out.append('replaytv.py - ')
        for x in data:
            debug_out.append(str(x))
        fdebug = open('debug.txt', 'a')
        fdebug.write(' '.join(debug_out))
        fdebug.close()
if hack83:
    debug_write(['Hack83 is enabled.\n'])

class Video(Plugin):

    CONTENT_TYPE = 'x-container/tivo-videos'

    # Used for 8.3's broken requests
    count = 0
    request_history = {}

    def send_file(self, handler, container, name):
        
        #No longer a 'cheep' hack :p
        if handler.headers.getheader('Range') and not handler.headers.getheader('Range') == 'bytes=0-':
            handler.send_response(206)
            handler.send_header('Connection', 'close')
            handler.send_header('Content-Type', 'video/x-tivo-mpeg')
            handler.send_header('Transfer-Encoding', 'chunked')
            handler.send_header('Server', 'TiVo Server/1.4.257.475')
            handler.end_headers()
            handler.wfile.write("\x30\x0D\x0A")
            return

        tsn =  handler.headers.getheader('tsn', '')

        o = urlparse("http://fake.host" + handler.path)
        path = unquote_plus(o[2])
        channelId = ""; showId = ""
        if len(path.split('/')) > 3:
           showId = path.split('/')[3]
           channelId = path.split('/')[2]
        print("send_File path = %s" % path)
        print ("channelId = %s" % channelId )
        print ("showId = %s" % showId )
        #see if we update the cache
        self.updateDvrShows()

        file_info = VideoDetails()
        file_info = self.getShowInfo(channelId, showId)
        handler.send_response(200)
        handler.end_headers()
        transcode.output_video(file_info['url'], file_info, handler.wfile, tsn)        

    def __isdir(self, full_path):
        return os.path.isdir(full_path)

    def __duration(self, full_path):
        return transcode.video_info(full_path)[4]

    def __est_size(self, duration, tsn = ''):
        #Size is estimated by taking audio and video bit rate adding 2%

        # Must be re-encoded
        audioBPS = strtod(config.getAudioBR(tsn))
        videoBPS = strtod(config.getVideoBR(tsn))
        bitrate =  audioBPS + videoBPS
        return int((duration/1000)*(bitrate * 1.02 / 8))
   
    def __getMetadataFromTxt(self, full_path):
        metadata = {}

        default_file = os.path.join(os.path.split(full_path)[0], 'default.txt')
        description_file = full_path + '.txt'

        metadata.update(self.__getMetadataFromFile(default_file))
        metadata.update(self.__getMetadataFromFile(description_file))

        return metadata

    def __getMetadataFromFile(self, file):
        metadata = {}

        if os.path.exists(file):
            for line in open(file):
                if line.strip().startswith('#'):
                    continue
                if not ':' in line:
                    continue

                key, value = line.split(':', 1)
                key = key.strip()
                value = value.strip()

                if key.startswith('v'):
                    if key in metadata:
                        metadata[key].append(value)
                    else:
                        metadata[key] = [value]
                else:
                    metadata[key] = value

        return metadata

    def __metadata(self, full_path):

        metadata = {}

        base_path, title = os.path.split(full_path)
        now = datetime.now()
        originalAirDate = datetime.fromtimestamp(os.stat(full_path).st_ctime)
        duration = self.__duration(full_path)
        duration_delta = timedelta(milliseconds = duration)
        
        metadata['title'] = '.'.join(title.split('.')[:-1])
        metadata['seriesTitle'] = metadata['title'] # default to the filename
        metadata['originalAirDate'] = originalAirDate.isoformat()
        metadata['time'] = now.isoformat()
        metadata['startTime'] = now.isoformat()
        metadata['stopTime'] = (now + duration_delta).isoformat()

        metadata.update( self.__getMetadataFromTxt(full_path) )
        
        metadata['size'] = self.__est_size(full_path)
        metadata['duration'] = duration

        min = duration_delta.seconds / 60
        sec = duration_delta.seconds % 60
        hours = min / 60
        min = min % 60
        metadata['iso_durarion'] = 'P' + str(duration_delta.days) + 'DT' + str(hours) + 'H' + str(min) + 'M' + str(sec) + 'S'

        return metadata

    def getMirrorURL(self):
        # Query tvdb for a list of mirrors
        mirrorsXML = minidom.parse(urllib.urlopen('http://thetvdb.com/interfaces/GetMirrors.php'))
        mirrors = mirrorsXML.getElementsByTagName("Item")
        mirrorURL = mirrors[0].getElementsByTagName("interface")[0]
        mirrorURL = self.getText(mirrorURL.childNodes)
        return mirrorURL
    ####
    def lookupSeriesId(self,mirrorURL,show_name,seriesCache):    
        # Get series information.. Assume first match is the best
        url = mirrorURL + GETSERIESID_URL + urllib.urlencode({"seriesname" : show_name})
        seriesXML = minidom.parse(urllib.urlopen(url))
        series = seriesXML.getElementsByTagName("Item")
        seriesid = "0111"
        if series:
          #TiVo seriesID is the "id", but return other details as well
          seriesidNode = series[0].getElementsByTagName("id")
          if seriesidNode:
            seriesid = self.getText(seriesidNode[0].childNodes)
        #todo get actors and genre as well
        #add series into the "cache"
        seriesCache[show_name] = seriesid
        return seriesid    
    #####
    def getSeriesId(self,show_name,seriesCache):
        if (seriesCache.has_key(show_name)):
            return seriesCache[show_name]
        else:
            return self.lookupSeriesId(mirrorURL,show_name,seriesCache)
    #####   
    def removeBlanks(self,list):
        newList = VideoDetails()
        for item in list.keys():
            value = list[item]
            if value != '':
                newList[item] = list[item]
        return newList
    #####
    def getText(self,nodelist):
        rc = ""
        for node in nodelist:
            if node.nodeType == node.TEXT_NODE:
                rc = rc + node.data
        return rc
    #####
    def getStreamingURL(self,show):
        streamingInfo = show.getElementsByTagName("STREAMING_INFO")[0]        
        return streamingInfo.getAttribute('URL')
    #####
    def getTVRatings(self,show):
        tvRatingXML = show.getElementsByTagName("TV_RATINGS")[0]
        tvRating = tvRatingXML.getAttribute('TV_RATING')
        #print '#Got tvRating - %s',tvRating
        if tvRating == 'TV-Y7':  tvRating = 'x1'
        elif  tvRating == 'TV-Y':  tvRating = 'x2'
        elif  tvRating == 'TV-G':  tvRating = 'x3'
        elif  tvRating == 'TV-PG':  tvRating = 'x4'
        elif  tvRating == 'TV-14':  tvRating = 'x5'
        elif  tvRating == 'TV-MA':  tvRating = 'x6'
        elif  tvRating == 'TV-NR':  tvRating = 'x7'
        else: tvRating = 'x0'
        return tvRating
    #####
    def getRecordingInfo(self,show):
        recordingInfo = show.getElementsByTagName("RECORDING_INFO")[0]
        quality = recordingInfo.getAttribute('RECORD_QUALITY') 
        recordedOn = recordingInfo.getAttribute('RECORDED_ON') 
        recordedAt = '%s:00' % recordingInfo.getAttribute('RECORDED_AT')
        #RECORDED_AT="20:00" 
        #RECORDED_ON="2007/11/15"

        captureDate = hex(int(time.mktime(time.strptime("%s %s" % (recordedOn, recordedAt), "%Y/%m/%d %H:%M:%S"))))

        temp = []
        #change date mapping replace / with -
        for i in range(0,len(recordedOn)): 
            if recordedOn[i] != '/': temp.append(recordedOn[i])
            else: temp.append('-')
        recordedOn = ''.join(temp)
        # add seconds to the recorded time
        # duration needs to be in millis so add 3 zeros to the end, subtract two
        duration = (int(recordingInfo.getAttribute('DURATION_IN_SECONDS'))-2)*1000

        return {
                'recordedOn': recordedOn,
                'recordedAt': recordedAt,
                'duration': duration,
                'captureDate' : captureDate,
                'quality': quality,
                'tvChannelNum': recordingInfo.getAttribute('TV_CHANNEL_NUM'),
                'tvChannelId' : recordingInfo.getAttribute('TV_CHANNEL_STATION_ID'),
                'repeat' : recordingInfo.getAttribute('IS_REPEAT')
            }
    #####
    def getShowInfo(self,channelId, showId):
        videos = dvrShowCache[channelId]
        for video in videos:
          if showId == video['id']: 
            return video
        return VideoDetails()          
    #####
    def getShow(self,show,seriesCache):
        recordingInfo = self.getRecordingInfo(show)
        video = VideoDetails()
        #todo
        #'originalAirDate',RECORDING_INFO RECORDED_ON+RECORDED_AT
        video['originalAirDate'] = recordingInfo['recordedOn'] + 'T' + recordingInfo['recordedAt'] + 'Z'
        video['time'] = video['originalAirDate']
        video['title'] = show.getAttribute('TITLE')
        video['id'] = show.getAttribute("ID")
        video['seriesTitle'] = video['title']
        video['episodeTitle'] = show.getAttribute('EPISODE')
        childTag = show.getElementsByTagName("EPISODE_SUMMARY")[0]
        video['description'] = self.getText(childTag.childNodes)
        if "" == video['description']: video['description'] = "n/a"
        #TODO Not quite working at the moment
        if "" != mirrorURL:
            video['seriesId'] = self.getSeriesId(video['title'],seriesCache)
        else:
            video['seriesId'] = "0111"
        if recordingInfo['repeat'] == 'TRUE': 
          video['showingBits'] = 0x00000200
        else: 
          video['showingBits'] = 0
        video['duration'] = recordingInfo['duration']
        video['captureDate'] = recordingInfo['captureDate']
        video['size'] = self.__est_size(int(video['duration']))
        video['quality'] = recordingInfo['quality']
        video['displayMajorNumber'] = recordingInfo['tvChannelNum']
        if "" == video['displayMajorNumber']: video['displayMajorNumber'] = "1"
        video['callsign'] = recordingInfo['tvChannelId']
        if "" == video['callsign']: video['callsign'] = "RTV"
        video['isEpisode'] = 'true'
        video['tvRating'] = self.getTVRatings(show)
        childTag = show.getElementsByTagName("ACTORS")[0]
        video['vActor'] = self.getText(childTag.childNodes).split(',')
        childTag = show.getElementsByTagName("GUEST_STARS")[0]
        video['vGuestStar'] = self.getText(childTag.childNodes).split(',')
        childTag = show.getElementsByTagName("DIRECTOR")[0]
        video['vDirector'] = self.getText(childTag.childNodes).split(',')
        childTag = show.getElementsByTagName("PRODUCERS")[0]
        video['vProducer'] = self.getText(childTag.childNodes).split(',')
        video['url'] = self.getStreamingURL(show)
        video['is_dir'] = 0    
        return self.removeBlanks(video)
        #return video
    #####
    def getShows(self,channel):
        seriesCache = {}
        shows = channel.getElementsByTagName("REPLAY_SHOW")        
        videos = []
        for show in shows:
            video = self.getShow(show,seriesCache)
            videos.append(video)
        return videos
    #####
    def getChannelsByDVR(self,dvr,dvrname):
        channels = dvr.getElementsByTagName("REPLAY_CHANNEL")        
        channelNames = []        
        for channel in channels:
            channelName = VideoDetails()
            channelName['name'] = channel.getAttribute("NAME")
            channelName['title'] = channel.getAttribute("NAME")
            channelName['id'] = channel.getAttribute("ID")
            channelName['uniquename'] = "%s_%s" % (dvrname, channelName['id'])
            channelName['is_dir'] = 1
            channelNames.append(channelName)
        return channelNames
    #####
    def getChannels(self,dvr):
        channels = dvr.getElementsByTagName("REPLAY_CHANNEL")        
        channelNames = []        
        for channel in channels:
            channelName = VideoDetails()
            channelName['name'] = channel.getAttribute("NAME")
            channelName['title'] = channel.getAttribute("NAME")
            channelName['id'] = channel.getAttribute("ID")
            channelName['uniquename'] = channel.getAttribute("NAME")
            channelName['is_dir'] = 1
            channelName['count'] = len(self.getShows(self.getChannel(dvr,channelName['id'])))
            channelNames.append(channelName)
        #sort channels by name
        def compareName(a,b): 
          return cmp(a['name'],b['name'])
        return sorted(channelNames,compareName)
    #####
    def getChannel(self,dvr,channelId):
        channels = dvr.getElementsByTagName("REPLAY_CHANNEL")
        for channel in channels:
            if channelId == channel.getAttribute("ID"):
                return channel
        return {}
    #####
    #Now cache by channelName, shows
    def updateDvrShows(self):
        #get current time less 1 hour to see if we need to update
        lastUpdateTime = dvrLastCacheTime[0]
        checkTime = datetime.now() - timedelta(hours=1)
        if lastUpdateTime < checkTime:
            #update cache
            print("Updating DVR Show cache")
            dvrShowCache.clear()
            replayXML = minidom.parse(urllib.urlopen(dvarchiveURL))
            channelList = self.getChannels(replayXML)
            dvrShowCache['channels'] = channelList
            for channelItem in channelList:
                channelName = channelItem['name']
                channelId = channelItem['id']
                #print("--channelName = %s - %s" % (channelName, channelId))
                channel = self.getChannel(replayXML,channelId)
                showList = self.getShows(channel)
                dvrShowCache["%s" % channelId] = showList 
            lastUpdateTime = datetime.now()
            dvrLastCacheTime[0] = lastUpdateTime
    #####
    def QueryContainer(self, handler, query):        
        tsn =  handler.headers.getheader('tsn', '')	
        subcname = query['Container'][0]           
        cname = subcname.split('/')[0]
        channelId = ""
        if len(subcname.split('/')) > 1:
           channelId = subcname.split('/')[1]
        videos = []
        uniquename = ""
        #update the cache if it has been over an hour
        self.updateDvrShows()
        if channelId == "":
            print("Loading Channel List")
            videos = dvrShowCache['channels']
            uniquename = "channels"
        else:
            print("Loading show List for channelId %s" % channelId)
            videos = dvrShowCache[channelId]     
            uniquename = channelId

        total = len(videos)
        #count limits the number of records returned default to 8
        count = 0
        if query.has_key('ItemCount'):
          count = int(query['ItemCount'][0])
        
        index = 0
        if total == total:
            if query.has_key('AnchorItem'):
              bs = '/TiVoConnect?Command=QueryContainer&Container='
              anchorItem = ""; anchorShowID = ""; anchorChannelId = "";
              anchorItem= query['AnchorItem'][0]
              if anchorItem.startswith(bs):
                  anchorItem= anchorItem.replace(bs, '/', 1)
              anchorItem= unquote(anchorItem)
              #print ("AnchorItem = %s" % anchorItem)
              if len(anchorItem.split('/')) > 3:
                 anchorShowID = anchorItem.split('/')[3]
                 anchorChannelId = anchorItem.split('/')[2]
                 index2 = 0
                 for video in videos:
                     if anchorShowID == video['id']: index = index2
                     index2 = index2 + 1
              elif len(anchorItem.split('/')) > 2:
                 anchorChannelId = anchorItem.split('/')[2]
                 index2 = 0
                 for video in videos:
                     if anchorChannelId == video['id']: index = index2
                     index2 = index2 + 1
              #print ("anchorShowID = %s" % anchorShowID)
              #print ("anchorChannelId = %s" % anchorChannelId)
              #print ("index = %s" % index)

              if count > 0:
                  index += 1
              if query.has_key('AnchorOffset'):
                index += int(query['AnchorOffset'][0])

              #foward count
              if count >= 0:
                  videos = videos[index:index + count]
              #backwards count
              else:
                  if index + count < 0:
                      count = -index
                  videos = videos[index + count:index]
                  index += count
            else:  # No AnchorItem
                if count >= 0:
                    videos = videos[:count]
                else:
                    index = count % len(videos)
                    videos = videos[count:]

        #todo handle links to details page
        handler.send_response(200)
        handler.end_headers()
        t = Template(file=os.path.join(SCRIPTDIR,'templates', 'container.tmpl'))
        t.container = cname
        t.name = subcname
        t.total = total
        t.start = index
        t.videos = videos
        t.quote = quote
        t.escape = escape
        t.uniquename = uniquename
        t.crc = zlib.crc32
        t.guid = config.getGUID()
        t.transcode = transcodeEnabled
        print("transcode = %s" % transcode)
        handler.wfile.write(t)


    def TVBusQuery(self, handler, query):       
        showId = query['File'][0]
        subcname = query['Container'][0]           
        cname = subcname.split('/')[0]
        channelId = ""
        if len(subcname.split('/')) > 1:
           channelId = subcname.split('/')[1]
        print("subcname = %s" % subcname)
        print ("channelId = %s" % channelId )
        print ("showId = %s" % showId )
        #see if we update the cache
        self.updateDvrShows()

        file_info = VideoDetails()
        file_info = self.getShowInfo(channelId, showId)
        #add on the extra fluff details
        duration = int(file_info['duration'])
        duration_delta = timedelta(milliseconds = duration)
        min = duration_delta.seconds / 60
        sec = duration_delta.seconds % 60
        hours = min / 60
        min = min % 60
        file_info['time'] = datetime.now().isoformat()
        file_info['startTime'] = datetime.now().isoformat()
        file_info['stopTime'] = (datetime.now() + duration_delta).isoformat()
        file_info['iso_duration'] = 'P' + str(duration_delta.days) + \
                                   'DT' + str(hours) + 'H' + str(min) + \
                                   'M' + str(sec) + 'S'

        handler.send_response(200)
        handler.end_headers()
        t = Template(file=os.path.join(SCRIPTDIR,'templates', 'TvBus.tmpl'))
        t.video = file_info
        t.escape = escape
        handler.wfile.write(t)
    
class VideoDetails(DictMixin):
   
    def __init__(self, d = None):
        if d:
            self.d = d
        else:
            self.d = {}

    def __getitem__(self, key):
        if key not in self.d:
            self.d[key] = self.default(key)
        return self.d[key]

    def __contains__(self, key):
        return True

    def __setitem__(self, key, value):
        self.d[key] = value

    def __delitem__(self):
        del self.d[key]
    
    def keys(self):
        return self.d.keys()
    
    def __iter__(self):
        return self.d.__iter__()

    def iteritems(self):
        return self.d.iteritems()

    def default(self, key):
        defaults = {
            'showingBits' : '0',
            'episodeNumber' : '0',
            'displayMajorNumber' : '0',
            'displayMinorNumber' : '0',
            'isEpisode' : 'true',
            'colorCode' : ('COLOR', '4'),
            'showType' : ('SERIES', '5'),
            'tvRating' : ('NR', '7'),
        }
        if key in defaults:
            return defaults[key]
        elif key.startswith('v'):
            return []
        else:
            return ''

        
# Parse a bitrate using the SI/IEEE suffix values as if by ffmpeg
# For example, 2K==2000, 2Ki==2048, 2MB==16000000, 2MiB==16777216
# Algorithm: http://svn.mplayerhq.hu/ffmpeg/trunk/libavcodec/eval.c
def strtod(value):
    prefixes = {"y":-24,"z":-21,"a":-18,"f":-15,"p":-12,"n":-9,"u":-6,"m":-3,"c":-2,"d":-1,"h":2,"k":3,"K":3,"M":6,"G":9,"T":12,"P":15,"E":18,"Z":21,"Y":24}
    p = re.compile(r'^(\d+)(?:([yzafpnumcdhkKMGTPEZY])(i)?)?([Bb])?$')
    m = p.match(value)
    if m is None:
        raise SyntaxError('Invalid bit value syntax')
    (coef, prefix, power, byte) = m.groups()
    if prefix is None:
        value = float(coef)
    else:
        exponent = float(prefixes[prefix])
        if power == "i":
            # Use powers of 2
            value = float(coef) * pow(2.0, exponent/0.3)
        else:
            # Use powers of 10
            value = float(coef) * pow(10.0, exponent)
    if byte == "B": # B==Byte, b=bit
        value *= 8;
    return value
