import subprocess, shutil, os, re, sys, ConfigParser, time, lrucache
import config

info_cache = lrucache.LRUCache(1000)


debug = config.getDebug()
MAX_VIDEO_BR = config.getMaxVideoBR()
BUFF_SIZE = config.getBuffSize()

FFMPEG = config.get('Server', 'ffmpeg')

def debug_write(data):
    if debug:
        debug_out = []
        for x in data:
            debug_out.append(str(x))
        fdebug = open('debug.txt', 'a')
        fdebug.write(' '.join(debug_out))
        fdebug.close()

# XXX BIG HACK
# subprocess is broken for me on windows so super hack
def patchSubprocess():
    o = subprocess.Popen._make_inheritable

    def _make_inheritable(self, handle):
        if not handle: return subprocess.GetCurrentProcess()
        return o(self, handle)

    subprocess.Popen._make_inheritable = _make_inheritable
mswindows = (sys.platform == "win32")
if mswindows:
    patchSubprocess()
        
#Hacked by RFryar to immediately transcode
def output_video(inFile, duration, outFile, tsn=''):
    transcode(inFile, duration, outFile, tsn)

def transcode(inFile, video_info, outFile, tsn=''):

    settings = {}
    #settings['audio_br'] = config.getAudioBR(tsn)
    # Need to hard code to 192
    settings['audio_br'] = '192k'

    settings['video_br'] = config.getVideoBR(tsn)
    settings['max_video_br'] = MAX_VIDEO_BR
    settings['buff_size'] = BUFF_SIZE
    #settings['aspect_ratio'] = ' ' .join(select_aspect(video_info, tsn))
    #commenting out aspect as it does not work correctly on LOW res videos.
    settings['aspect_ratio'] = ' ' #.join(select_aspect(video_info, tsn))
	
    #NEw settings here 
    settings['audio_codec'] = '-acodec mp2 -ac 2'
    settings['audio_fr'] = '-ar 48000'
    settings['video_fps'] = '-r 29.97'


    cmd_string = config.getFFMPEGTemplate(tsn) % settings
    #cmd_string = " -sameq -aspect 4:3 -flags ildct+ilme -t %s %s" % (video_info['duration'], cmd_string)
    #test adding sameq to the command string SLOW
    #test forcing interlaced on source, did not fix standard motion issues
    #cmd_string = " -top 1 -flags ildct+ilme %s" % cmd_string
    #try to deinterlace
    cmd_string = " -deinterlace %s" % cmd_string
    print("cmd_string2 %s" % cmd_string)

    cmd = [FFMPEG, '-i', inFile] + cmd_string.split()

    debug_write(['transcode: ffmpeg command is ', ' '.join(cmd), '\n'])
    ffmpeg = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    try:
        shutil.copyfileobj(ffmpeg.stdout, outFile)
    except:
        kill(ffmpeg.pid)
       
def select_aspect(video, tsn = ''):
    TIVO_WIDTH = config.getTivoWidth(tsn)
    TIVO_HEIGHT = config.getTivoHeight(tsn)
    
    inFile = video['url']
    millisecs = video['duration']    
    quality = video['quality']    
    width = 720; height = 480; fps = "29.97"; type = "mpeg2video"
    #'HIGH' == quality:
    #  720x480, vid br- 9800, aud br-192, 29.97 tb(r)
    #'MEDIUM' == quality
    #  720x480, vid br- 7413, aud br-192, 29.97 tb(r)
    if 'LOW' == quality:
        #  352x480, vid br- 4004, aud br-192, 29.97 tb(r)
        width = 352

    debug_write(['tsn:', tsn, '\n'])

    aspect169 = config.get169Setting(tsn)

    debug_write(['aspect169:', aspect169, '\n'])

    optres = config.getOptres()

    debug_write(['optres:', optres, '\n'])

    if optres:
        optHeight = config.nearestTivoHeight(height)
        optWidth = config.nearestTivoWidth(width)
        if optHeight < TIVO_HEIGHT:
            TIVO_HEIGHT = optHeight
        if optWidth < TIVO_WIDTH:
            TIVO_WIDTH = optWidth

    d = gcd(height,width)
    ratio = (width*100)/height
    rheight, rwidth = height/d, width/d

    debug_write(['select_aspect: File=', inFile, ' Type=', type, ' width=', width, ' height=', height, ' fps=', fps, ' millisecs=', millisecs, ' ratio=', ratio, ' rheight=', rheight, ' rwidth=', rwidth, ' TIVO_HEIGHT=', TIVO_HEIGHT, 'TIVO_WIDTH=', TIVO_WIDTH, '\n'])

    multiplier16by9 = (16.0 * TIVO_HEIGHT) / (9.0 * TIVO_WIDTH)
    multiplier4by3  =  (4.0 * TIVO_HEIGHT) / (3.0 * TIVO_WIDTH)
   
    if (rwidth, rheight) in [(4, 3), (10, 11), (15, 11), (59, 54), (59, 72), (59, 36), (59, 54)]:
        debug_write(['select_aspect: File is within 4:3 list.\n'])
        return ['-aspect', '4:3', '-s', str(TIVO_WIDTH) + 'x' + str(TIVO_HEIGHT)]
    elif ((rwidth, rheight) in [(16, 9), (20, 11), (40, 33), (118, 81), (59, 27)]) and aspect169:
        debug_write(['select_aspect: File is within 16:9 list and 16:9 allowed.\n'])
        return ['-aspect', '16:9', '-s', str(TIVO_WIDTH) + 'x' + str(TIVO_HEIGHT)]
    else:
        settings = []
        #If video is wider than 4:3 add top and bottom padding
        if (ratio > 133): #Might be 16:9 file, or just need padding on top and bottom
            if aspect169 and (ratio > 135): #If file would fall in 4:3 assume it is supposed to be 4:3 
                if (ratio > 177):#too short needs padding top and bottom
                    endHeight = int(((TIVO_WIDTH*height)/width) * multiplier16by9)
                    settings.append('-aspect')
                    settings.append('16:9')
                    if endHeight % 2:
                        endHeight -= 1
                    if endHeight < TIVO_HEIGHT * 0.99:
                        settings.append('-s')
                        settings.append(str(TIVO_WIDTH) + 'x' + str(endHeight))

                        topPadding = ((TIVO_HEIGHT - endHeight)/2)
                        if topPadding % 2:
                            topPadding -= 1
                        
                        settings.append('-padtop')
                        settings.append(str(topPadding))
                        bottomPadding = (TIVO_HEIGHT - endHeight) - topPadding
                        settings.append('-padbottom')
                        settings.append(str(bottomPadding))
                    else:   #if only very small amount of padding needed, then just stretch it
                        settings.append('-s')
                        settings.append(str(TIVO_WIDTH) + 'x' + str(TIVO_HEIGHT))
                    debug_write(['select_aspect: 16:9 aspect allowed, file is wider than 16:9 padding top and bottom\n', ' '.join(settings), '\n'])
                else: #too skinny needs padding on left and right.
                    endWidth = int((TIVO_HEIGHT*width)/(height*multiplier16by9))
                    settings.append('-aspect')
                    settings.append('16:9')
                    if endWidth % 2:
                        endWidth -= 1
                    if endWidth < (TIVO_WIDTH-10):
                        settings.append('-s')
                        settings.append(str(endWidth) + 'x' + str(TIVO_HEIGHT))

                        leftPadding = ((TIVO_WIDTH - endWidth)/2)
                        if leftPadding % 2:
                            leftPadding -= 1

                        settings.append('-padleft')
                        settings.append(str(leftPadding))
                        rightPadding = (TIVO_WIDTH - endWidth) - leftPadding
                        settings.append('-padright')
                        settings.append(str(rightPadding))
                    else: #if only very small amount of padding needed, then just stretch it
                        settings.append('-s')
                        settings.append(str(TIVO_WIDTH) + 'x' + str(TIVO_HEIGHT))
                    debug_write(['select_aspect: 16:9 aspect allowed, file is narrower than 16:9 padding left and right\n', ' '.join(settings), '\n'])
            else: #this is a 4:3 file or 16:9 output not allowed
                settings.append('-aspect')
                settings.append('4:3')
                endHeight = int(((TIVO_WIDTH*height)/width) * multiplier4by3)
                if endHeight % 2:
                    endHeight -= 1
                if endHeight < TIVO_HEIGHT * 0.99:
                    settings.append('-s')
                    settings.append(str(TIVO_WIDTH) + 'x' + str(endHeight))

                    topPadding = ((TIVO_HEIGHT - endHeight)/2)
                    if topPadding % 2:
                        topPadding -= 1
                    
                    settings.append('-padtop')
                    settings.append(str(topPadding))
                    bottomPadding = (TIVO_HEIGHT - endHeight) - topPadding
                    settings.append('-padbottom')
                    settings.append(str(bottomPadding))
                else:   #if only very small amount of padding needed, then just stretch it
                    settings.append('-s')
                    settings.append(str(TIVO_WIDTH) + 'x' + str(TIVO_HEIGHT))
                debug_write(['select_aspect: File is wider than 4:3 padding top and bottom\n', ' '.join(settings), '\n'])

            return settings
        #If video is taller than 4:3 add left and right padding, this is rare. All of these files will always be sent in
        #an aspect ratio of 4:3 since they are so narrow.
        else:
            endWidth = int((TIVO_HEIGHT*width)/(height*multiplier4by3))
            settings.append('-aspect')
            settings.append('4:3')
            if endWidth % 2:
                endWidth -= 1
            if endWidth < (TIVO_WIDTH * 0.99):
                settings.append('-s')
                settings.append(str(endWidth) + 'x' + str(TIVO_HEIGHT))

                leftPadding = ((TIVO_WIDTH - endWidth)/2)
                if leftPadding % 2:
                    leftPadding -= 1

                settings.append('-padleft')
                settings.append(str(leftPadding))
                rightPadding = (TIVO_WIDTH - endWidth) - leftPadding
                settings.append('-padright')
                settings.append(str(rightPadding))
            else: #if only very small amount of padding needed, then just stretch it
                settings.append('-s')
                settings.append(str(TIVO_WIDTH) + 'x' + str(TIVO_HEIGHT))

            debug_write(['select_aspect: File is taller than 4:3 padding left and right\n', ' '.join(settings), '\n'])
            
            return settings

#not used
def tivo_compatable(inFile, tsn = ''):
    print "In unused method: tivo_compatable!"
    return True

#not used
def video_info(inFile):
    print "In unused method: video_info!"
    return "mpeg2video", 720, 480, "29.97", 10000

#not used       
def supported_format(inFile):
    print "In unused method: supported_format!"
    return True

def kill(pid):
    debug_write(['kill: killing pid=', str(pid), '\n'])
    if mswindows:
        win32kill(pid)
    else:
        import os, signal
        os.kill(pid, signal.SIGTERM)

def win32kill(pid):
        import ctypes
        handle = ctypes.windll.kernel32.OpenProcess(1, False, pid)
        ctypes.windll.kernel32.TerminateProcess(handle, -1)
        ctypes.windll.kernel32.CloseHandle(handle)

def gcd(a,b):
    while b:
        a, b = b, a % b
    return a
